#include <sfc-performance/sfc.hpp>

namespace SuperFamicom {

PPU ppu;
#include "io.cpp"
#include <sfc/ppu-fast/line.cpp>
#include <sfc/ppu-fast/background.cpp>
#include <sfc/ppu-fast/mode7.cpp>
#include <sfc/ppu-fast/object.cpp>
#include <sfc/ppu-fast/window.cpp>
#include <sfc/ppu-fast/serialization.cpp>
#include <sfc/ppu/counter/serialization.cpp>

PPU::PPU() {
  output = new uint32[512 * 512] + 16 * 512;  //overscan offset
  #if defined(DOUBLE_VRAM)
  tilecache[TileMode::BPP2] = new uint8[8192 * 8 * 8];
  tilecache[TileMode::BPP4] = new uint8[4096 * 8 * 8];
  tilecache[TileMode::BPP8] = new uint8[2048 * 8 * 8];
  #else
  tilecache[TileMode::BPP2] = new uint8[4096 * 8 * 8];
  tilecache[TileMode::BPP4] = new uint8[2048 * 8 * 8];
  tilecache[TileMode::BPP8] = new uint8[1024 * 8 * 8];
  #endif

  for(uint y : range(lines.size())) {
    lines[y].y = y;
  }

  for(uint layer : range(4)) for(uint priority : range(2)) BGEnabled[layer][priority] = true;
  for(uint priority : range(4)) OBJEnabled[priority] = true;

  FrameSkip = 0;
  FrameCounter = 0;
}

PPU::~PPU() {
  delete[] (output - 16 * 512);  //overscan offset
  delete[] tilecache[TileMode::BPP2];
  delete[] tilecache[TileMode::BPP4];
  delete[] tilecache[TileMode::BPP8];
}

auto PPU::Enter() -> void {
  while(true) scheduler.synchronize(), ppu.main();
}

auto PPU::step(uint clocks) -> void {
  tick(clocks);
  Thread::step(clocks);
  synchronize(cpu);
}

auto PPU::main() -> void {
  if(FrameCounter > 0) {
    if(vcounter() == 0) {
      scheduler.exit(Scheduler::Event::Frame);
      FrameCounter = FrameSkip == 0 ? 0 : (FrameCounter + 1) % FrameSkip;
    }
    step(lineclocks() - hcounter());
    return;
  }
  scanline();
  uint y = vcounter();
  step(512);
  if(y >= 1 && y <= 239) {
    if(io.displayDisable || y >= vdisp()) {
      lines[y].io.displayDisable = true;
    } else {
      memcpy(&lines[y].io, &io, sizeof(io));
      memcpy(&lines[y].cgram, &cgram, sizeof(cgram));
    }
    if(!Line::count) Line::start = y;
    Line::count++;
  }
  if(FrameSkip > 1 && vcounter() == 0) FrameCounter += 1;
  step(lineclocks() - hcounter());
}

auto PPU::scanline() -> void {
  if(vcounter() == 0) {
    display.interlace = io.interlace;
    display.overscan = io.overscan;
    latch.hires = false;
    io.obj.timeOver = false;
    io.obj.rangeOver = false;
  }

  if(vcounter() > 0 && vcounter() < vdisp()) {
    latch.hires |= io.pseudoHires || io.bgMode == 5 || io.bgMode == 6;
    latch.hires |= io.bgMode == 7 && configuration.hacks.ppuFast.hiresMode7;
  }

  if(vcounter() == vdisp() && !io.displayDisable) {
    oamAddressReset();
  }

  if(vcounter() == 240) {
    Line::flush();
    scheduler.exit(Scheduler::Event::Frame);
  }
}

auto PPU::refresh() -> void {
  auto output = this->output;
  if(!overscan()) output -= 14 * 512;
  auto pitch  = 512 << !interlace();
  auto width  = 256 << hires();
  auto height = 240 << interlace();
  Emulator::video.setEffect(Emulator::Video::Effect::ColorBleed, configuration.video.blurEmulation && hires());
  Emulator::video.refresh(output, pitch * sizeof(uint32), width, height);
}

auto PPU::load() -> bool {
  ppu1.version = max(1, min(1, configuration.system.ppu1.version));
  ppu2.version = max(1, min(3, configuration.system.ppu2.version));
  #if defined(DOUBLE_VRAM)
  vram.mask = configuration.system.ppu1.vram.size / sizeof(uint16) - 1;
  if(vram.mask != 0xffff) vram.mask = 0x7fff;
  #endif
  return true;
}

auto PPU::power(bool reset) -> void {
  create(Enter, system.cpuFrequency());
  PPUcounter::reset();
  memory::fill<uint32>(output, 512 * 480);

  function<auto (uint24, uint8) -> uint8> reader{&PPU::readIO, this};
  function<auto (uint24, uint8) -> void> writer{&PPU::writeIO, this};
  bus.map(reader, writer, "00-3f,80-bf:2100-213f");

  if(!reset) {
    for(auto address : range(vram.mask + 1)) {
      vram[address] = 0x0000;
      updateTiledata(address);
    }
    for(auto& color : cgram) color = 0x0000;
    for(auto& object : objects) object = {};
  }

  latch = {};
  io = {};
  updateVideoMode();

  ItemLimit = !configuration.hacks.ppuFast.noSpriteLimit ? 32 : 128;
  TileLimit = !configuration.hacks.ppuFast.noSpriteLimit ? 34 : 128;

  Line::start = 0;
  Line::count = 0;
}

auto PPU::setLayerEnabled(uint layer, uint priority, bool enable) -> void {
  if(layer < 4) {
    BGEnabled[layer][priority] = enable;
  } else {
    OBJEnabled[priority] = enable;
  }
}

auto PPU::setFrameSkip(uint frameskip) -> void {
  FrameSkip = frameskip;
  FrameCounter = 0;
}

}
